from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import ciphers, hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa, padding
from cryptography.hazmat.primitives.ciphers import algorithms, modes
from cryptography.exceptions import InvalidSignature


from getpass import getpass
from uuid import uuid4
import binascii
import logging
import base64
import stat
import sys
import io
import os


class SecurityError(ValueError):
    pass

def is_key_string(s):
    if s.startswith('-----BEGIN'):
        return True
    if s.startswith('ssh-rsa'):
        return True
    #TODO: Support other formats
    return False

def load_rsa_key(source, passphrase=None, require_private=True):
    try:
        if is_key_string(source):
            key_contents = source
        elif require_private:
            key_contents = read_private_path(source)
        else:
            key_contents = open(source).read()
    except FileNotFoundError:
        return None
    if 'ENCRYPTED' in key_contents:
        if passphrase is None:
            passphrase = getpass('Enter the passphrase for "{}": '.format(source))
    return RSA(key_contents, passphrase)

def generate_uuid():
    return str(uuid4())

def hasher(data, raw=False):
    if isinstance(data, str):
        data = data.encode('utf-8')

    hasher = hashes.Hash(hashes.SHA256(), backend=default_backend())
    hasher.update(data)
    digest = hasher.finalize()
    if raw:
        return digest
    else:
        return binascii.hexlify(digest).decode('utf-8')

def generate_random_bytes(size=64):
    return os.urandom(size)

def generate_int(start=0, end=sys.maxsize):
    """ inclusive to exclusive """
    delta = end - start
    random_int = int.from_bytes(generate_random_bytes(20), byteorder="big")
    number = (random_int % delta) + start
    return number

def random_choice(items):
    index = generate_int(end=len(items))
    return items[index]

def generate_secret_key(length=128):
    random = generate_random_bytes(length)
    key = b64encode(random).decode('utf-8')
    return key

def to_b64_str(the_bytes, encoding='utf-8'):
    return base64.b64encode(the_bytes).decode(encoding)

def from_b64_str(string, encoding='utf-8'):
    return base64.b64decode(string.encode(encoding))

class AES():
    name = 'AES'
    cipher_name = 'AES::CFB'
    block_size = 16 # In bytes; so 16 is a 128-bit key
    default_key_size = block_size

    @classmethod
    def generate_key(cls, key_size=None):
        if key_size is None:
            key_size = cls.default_key_size
        key = generate_random_bytes(key_size)
        return to_b64_str(key)

    def __init__(self, key=None):
        if key is None:
            key = self.generate_key()
        self.set_key(key)

    def set_key(self, key):
        if isinstance(key, str):
            self.key = from_b64_str(key)
        else:
            self.key = key

    def get_key(self):
        return to_b64_str(self.key)

    def _cipher(self, iv):
        return ciphers.Cipher(
            algorithms.AES(self.key),
            modes.CFB(iv),
            backend=default_backend()
        )

    def encrypt(self, payload):
        iv = generate_random_bytes(self.block_size)
        cipher = ciphers.Cipher(
            algorithms.AES(self.key),
            modes.CFB(iv),
            backend=default_backend(),
        )
        encryptor = cipher.encryptor()
        ciphertext = encryptor.update(payload) + encryptor.finalize()
        metadata = {}
        metadata['cipher'] = self.name
        metadata['key_size'] = len(self.key) * 8
        metadata['iv'] = to_b64_str(iv)
        metadata['mode'] = 'CFB'
        return ciphertext, metadata

    def decrypt(self, ciphertext, iv):
        iv = from_b64_str(iv)
        cipher = self._cipher(iv)
        decryptor = cipher.decryptor()
        payload = decryptor.update(ciphertext) + decryptor.finalize()
        return payload

class RSA():
    name = 'RSA'
    cipher_name = 'RSAES-OAEP::SHA256'
    default_key_size = 2048
    public_exponent = 65537

    def __init__(self, key=None, passphrase=None):
        self._private = None
        self._public = None
        self._from_string(key, passphrase=passphrase)

    @classmethod
    def generate_key(cls, size=None):
        if not size:
            size = cls.default_key_size
        private_key = rsa.generate_private_key(
            public_exponent=cls.public_exponent,
            key_size=size,
            backend=default_backend(),
        )
        exported = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        ).decode('utf-8')
        return exported

    def _from_string(self, key, passphrase=None):
        key_bytes = key.encode('utf-8')
        if key.startswith('ssh-rsa'):
            self._public = serialization.load_ssh_public_key(
                key_bytes,
                backend=default_backend(),
            )
        elif 'PUBLIC' in key: # Is there a better test? pro detection for sure
            self._public = serialization.load_pem_public_key(
                key_bytes,
                backend=default_backend(),
            )
        else:
            if passphrase:
                passphrase = passphrase.encode('utf-8')
            self._private = serialization.load_pem_private_key(
                key_bytes,
                password=passphrase,
                backend=default_backend(),
            )

    def _get_public_key(self):
        if not self._public:
            if not self._private:
                raise ValueError('Unable to get a public key, no key information')
            self._public = self._private.public_key()
        return self._public

    def _get_private_key(self):
        if not self._private:
            raise ValueError('Private key information unavailable')
        return self._private

    def has_private_portion(self):
        return bool(self._private)

    def export_private_key(self, passphrase=None):
        options = {
            'encoding': serialization.Encoding.PEM,
            'format': serialization.PrivateFormat.PKCS8,
        }
        if passphrase:
            options['encryption_algorithm'] = serialization.BestAvailableEncryption(
                passphrase.encode('utf-8'),
            )
        else:
            options['encryption_algorithm'] = serialization.NoEncryption()
        exported = self._get_private_key().private_bytes(**options).decode('utf-8')
        return exported

    def export_public_key(self, format='pem'):
        if format == 'pem':
            encoding = serialization.Encoding.PEM
            key_format = serialization.PublicFormat.SubjectPublicKeyInfo
        elif format == 'openssh':
            encoding = serialization.Encoding.OpenSSH
            key_format = serialization.PublicFormat.OpenSSH

        public_key = self._get_public_key()
        exported = public_key.public_bytes(
            format=key_format,
            encoding=encoding,
        ).decode('utf-8')
        return exported

    def get_fingerprint(self, size=32):
        public_key = self._get_public_key()
        exported = public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )
        digest = hasher(exported)[:size]
        parts = [digest[i * 2:i * 2 + 2] for i in range(size // 2)] 
        fingerprint = ':'.join(parts)
        return fingerprint

    def sign(self, data):
        if isinstance(data, str):
            data = data.encode('utf-8')
        signer = self._get_private_key().signer(
            padding.PSS(
                mgf=padding.MGF1(hashes.SHA256()),
                salt_length=padding.PSS.MAX_LENGTH
            ),
            hashes.SHA256()
        )
        signer.update(data)
        signature = signer.finalize()
        return to_b64_str(signature)

    def verify(self, data, signature):
        if isinstance(data, str):
            data = data.encode('utf-8')
        if isinstance(signature, str):
            signature = from_b64_str(signature)
        public_key = self._get_public_key()
        verifier = public_key.verifier(
            signature,
            padding.PSS(
                mgf=padding.MGF1(hashes.SHA256()),
                salt_length=padding.PSS.MAX_LENGTH
            ),
            hashes.SHA256()
        )
        verifier.update(data)
        try:
            verifier.verify()
        except InvalidSignature:
            return False
        return True

    def encrypt(self, payload):
        public_key = self._get_public_key()
        ciphertext = public_key.encrypt(
            payload,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None
            )
        )
        metadata = {}
        metadata['cipher'] = self.cipher_name
        metadata['padding'] = 'OAEP'
        return ciphertext, metadata

    def decrypt(self, ciphertext, metadata=None):
        plaintext = self._get_private_key().decrypt(
            ciphertext,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None
            )
        )
        return plaintext

PRIVATE_FILE_MODE = stat.S_IRUSR | stat.S_IWUSR  # This is 0o600 in octal
PRIVATE_DIR_MODE = stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR  # This is 0o700 in octal

def make_private_dir(path):
    os.makedirs(path, mode=PRIVATE_DIR_MODE, exist_ok=False)

def remove_path(path, is_dir=False):
    path = os.path.expanduser(path)
    if is_dir:
        shutil.rmtree(path)
    else:
        os.remove(path)

def write_private_path(path, contents, mode='w', makedirs=True):
    path = os.path.expanduser(path)
    flags = os.O_WRONLY | os.O_CREAT | os.O_EXCL  # Refer to "man 2 open".
    if makedirs:
        directory = os.path.dirname(path)
        try:
            make_private_dir(directory)
        except FileExistsError:
            pass

    # For security, remove file with potentially elevated mode
    try:
        remove_path(path)
    except OSError:
        pass

    # Open file descriptor
    umask_original = os.umask(0)
    try:
        descriptor = os.open(path, flags, PRIVATE_FILE_MODE)
    finally:
        os.umask(umask_original)

    # Open file handle and write to file
    with os.fdopen(descriptor, mode) as file_writer:
        file_writer.write(contents)

def read_private_path(path, open_mode='r'):
    """
        Technically there is a race condition here between
        when im checking the mode and when I'm opening the file
        that I'm not sure I can get rid of. But it is private
        at the time of checking it, presumably it would require
        the authorization of the user to change it before being
        read.
    """
    path = os.path.expanduser(path)
    file_mode = stat.S_IMODE(os.stat(path).st_mode)
    link_mode = stat.S_IMODE(os.lstat(path).st_mode)
    if file_mode != PRIVATE_FILE_MODE or link_mode != PRIVATE_FILE_MODE: 
        logging.error('File has mode {} and needs {} to be considered private'.format(
            oct(file_mode), oct(PRIVATE_FILE_MODE)
        ))
        raise SecurityError('SECURITY ERROR: Unable to load file: {}. Insecure file permissions!'.format(path))
    with open(path, open_mode) as f:
        data = f.read()
    return data



import struct
import json
import io

class EncryptionInterchangeFile():
    HEADER_TITLE = b'encx' 
    METADATA_ENCODING = 'utf-8'

    def __init__(self, metadata=None, payload=None):
        self.metadata = metadata
        self.payload = payload

    def _parse_header(self, stream):
        title_bytes = stream.read(len(self.HEADER_TITLE))
        if title_bytes != self.HEADER_TITLE:
            raise ValueError('File does not appear to be the right format')

        metadata_size_bytes = struct.unpack('<L', stream.read(4))[0]

        metadata_bytes = stream.read(metadata_size_bytes)
        json_metadata = metadata_bytes.decode(self.METADATA_ENCODING)
        return json.loads(json_metadata)

    def _create_header(self, stream, metadata):
        encoded_metadata = json.dumps(metadata).encode('utf-8')
        stream.write(self.HEADER_TITLE)
        stream.write(struct.pack('<L', len(encoded_metadata)))
        stream.write(encoded_metadata)
        return stream

    def to_file(self, target):
        if isinstance(target, str):
            with open(target, 'wb') as f:
                self._create_header(f, self.metadata)
                target.write(self.payload.read())
        else:
            self._create_header(target, self.metadata)
            target.write(self.payload.read())
        return target

    def load_file(self, target):
        if isinstance(target, str):
            with open(target, 'rb') as f:
                self.metadata = self._parse_header(f)
                self.payload = io.BytesIO(f.read())
        else:
            self.metadata = self._parse_header(target)
            self.payload = io.BytesIO(target.read())

    @classmethod
    def from_file(cls, target):
        new_obj = cls()
        new_obj.load_file(target)
        return new_obj

ENCX = EncryptionInterchangeFile



from . import security
import logging

class KeyAliasNotFoundError(KeyError):
    pass

class MultipleKeysFoundError(KeyError):
    pass

class IncorrectKeyTypeError(ValueError):
    pass

class InvalidAliasError(ValueError):
    pass

class KeyStore():
    """
        Stores private key paths (no actual data), Aliases public
        and private keys.
    """
    def __init__(self, data={}):
        self.load_data(data)
        self._changed = False

    def load_data(self, data):
        if not data:
            data = {}
        self.data = data

    def has_changed(self):
        return self._changed

    def mark_changed(self):
        self._changed = True

    def export(self):
        return self.data

    def has_key(self, name):
        if name in self.data:
            return True
        return False

    def is_trusted_key(self, public_key):
        openssh_key = public_key.export_public_key('openssh')
        for name, entry in self.data.items():

            if entry.get('type') == 'private_key':
                # We're only interested in actual keys
                entry_value = entry.get('public_key')
            elif entry.get('type') == 'public_key':
                entry_value = entry.get('value')
            else:
                # We're only interested in actual keys
                continue

            if entry_value == openssh_key:
                logging.info('Confirming key is trusted with name: {}.'.format(name))
                return True
        return False

    def delete_key(self, name):
        self.data.pop(name, None)
        self.mark_changed()

    def add_private_key(self, name, path, validate=True):
        key = security.load_rsa_key(path)
        assert key.has_private_portion()
        self.mark_changed()
        self.data[name] = {
            'type': 'private_key',
            'value': path,
            'public_key': key.export_public_key('openssh'),
        }

    def add_public_key(self, name, key):
        self.mark_changed()
        self.data[name] = {
            'type': 'public_key',
            'value': key.export_public_key('openssh'),
        }

    def add_alias(self, alias, names):
        self.mark_changed()
        self.data[alias] = {
            'type': 'alias',
            'value': names,
        }

    def add_to_alias(self, alias, names):
        entry = self.data.get(alias, None)
        if not entry or entry['type'] != 'alias':
            raise InvalidAliasError('Cannot add to an alias that doesnt exist')
        self.mark_changed()
        entry['value'].extend(names)

    def resolve_alias(self, root_alias):
        matches = []
        aliases = [root_alias]
        seen_aliases = {root_alias: True}

        # Resolve to entries
        while aliases:
            # Map to values
            next_aliases = []
            for a in aliases:
                entry = self.data.get(a, None)
                if not entry:
                    logging.warn('Key alias "{}" does not exist.'.format(a))
                    continue

                value = entry['value']
                if entry['type'] == 'alias':
                    for new_alias in value:
                        if new_alias in seen_aliases:
                            # Skip any we've already come across
                            # This prevents infinite looping
                            continue
                        next_aliases.append(new_alias)
                        seen_aliases[new_alias] = True
                else:
                    matches.append(entry)
            aliases = next_aliases

        # Entries to KeyObj
        return matches

    def _entries_to_keys(self, entries):
        """
            Note: This could be optimized to return only the portion needed
            and when the cached public portion is returned would save the
            time necessary to derive it from the private portion.
            Not sure if I want this optimization done here or if it is a benefit
            to have the flexibilty elsewhere to do multiple things with it.
        """
        keys = []
        for entry in entries:
            if entry['type'] == 'private_key':
                keys.append(security.load_rsa_key(entry['value']))
            else:
                keys.append(security.RSA(entry['value']))
        return keys

    def get_private_key(self, alias, require_match=True):
        matches = self.resolve_alias(alias)
        if not matches and require_match:
            raise KeyAliasNotFoundError('Alias {} not found!'.format(alias))
        if len(matches) > 1:
            raise MultipleKeysFoundError('Alias {} returned multiple entries!'.format(alias))
        entry = matches[0]
        if not entry:
            return None
        if entry['type'] != 'private_key':
            raise IncorrectKeyTypeError('Alias {} returned a public key!'.format(alias))
        return self._entries_to_keys(matches)[0]

    def get_public_keys(self, alias, require_match=True):
        matches = self.resolve_alias(alias)
        if not matches and require_match:
            raise KeyAliasNotFoundError('Alias {} not found!'.format(alias))
        return self._entries_to_keys(matches)

    def get_public_key(self, alias, require_match=True):
        matches = self.get_public_keys(alias, require_match=require_match)
        if not len(matches) == 1:
            raise MultipleKeysFoundError('Alias {} returned multiple entries!'.format(alias))
        return self._entries_to_keys(matches)[0]

    def get_entries(self, alias):
        return self.resolve_alias(alias)



import unittest
from unittest import mock

import base64
import yaml
import json

import io


from .spec import ENCX
from .security import (
    generate_random_bytes,
    to_b64_str, from_b64_str,
    AES, RSA
)
from . import security
from .keystore import KeyStore
from .schemes import all_schemes
from . import commands

class UtilityTests(unittest.TestCase):
    def test_random_bytes(self):
        bytes_1 = generate_random_bytes(16)
        self.assertEqual(len(bytes_1), 16)

        bytes_2 = generate_random_bytes(32)
        self.assertEqual(len(bytes_2), 32)

        bytes_3 = generate_random_bytes(32)
        self.assertEqual(len(bytes_3), 32)

        self.assertNotEqual(bytes_2, bytes_3)

    def test_b64_strings(self):
        value = generate_random_bytes(16)

        str_value = to_b64_str(value)
        there_and_back_again = from_b64_str(str_value)
        self.assertEqual(value, there_and_back_again)

class EncryptionSchemeTests(unittest.TestCase):
    def test_schemes(self):
        for Scheme in all_schemes:
            my_value = generate_random_bytes(100)

            metadata = {}

            # Encrypt our value
            key = Scheme.generate_key()
            enc_scheme = Scheme([key])
            ciphertext, meta = enc_scheme.encrypt(my_value)

            # ... and back again
            dec_scheme = Scheme([key])
            payload = dec_scheme.decrypt(ciphertext, meta)

            self.assertEqual(payload, my_value)

class FileFormatTest(unittest.TestCase):
    def test_basic(self):
        metadata = {'foo': 'bar', 'dataz': 42}
        my_bytes = generate_random_bytes(100)

        my_fake_file = io.BytesIO()

        ex = ENCX(metadata, io.BytesIO(my_bytes))
        ex.to_file(my_fake_file)

        my_fake_file.seek(0)

        reloaded = ENCX.from_file(my_fake_file)

        assert reloaded.metadata == metadata
        assert reloaded.payload.read() == my_bytes

class KeyStoreTests(unittest.TestCase):
    def test_add_private_key(self):
        path = '/tmp/encx-tests-1.pem'
        pem_private_1 = RSA.generate_key()
        with mock.patch.object(
            security,
            'read_private_path',
            return_value=pem_private_1
        ):
            key_store = KeyStore()
            key_store.add_private_key('john', path)
            reloaded_key = key_store.get_private_key('john')
            self.assertTrue(reloaded_key.export_private_key() == pem_private_1)

    def test_save(self):
        first_store = KeyStore()
        path = '/tmp/encx-tests-1.pem'
        pem_private_1 = RSA.generate_key()
        with mock.patch.object(
            security,
            'read_private_path',
            return_value=pem_private_1
        ):
            self.assertFalse(first_store.has_changed())
            first_store.add_private_key('john', path)
            self.assertTrue(first_store.has_changed())

            # Save and reload
            save = first_store.export()
            reloaded = KeyStore(save)
            reloaded_key = reloaded.get_private_key('john')
            self.assertTrue(reloaded_key.export_private_key() == pem_private_1)

    def test_add_public_key(self):
        generated_key = RSA(RSA.generate_key())
        generated_pub = generated_key.export_public_key('openssh')

        key_store = KeyStore()
        key_store.add_public_key('john', generated_key)
        match = key_store.get_public_key('john')
        self.assertTrue(generated_pub == match.export_public_key('openssh'))

    def test_aliases(self):
        generated_key = RSA(RSA.generate_key())
        generated_pub = generated_key.export_public_key('openssh')

        key_store = KeyStore()
        key_store.add_public_key('john', generated_key)
        key_store.add_alias('A_1', ['john'])
        key_store.add_alias('A_2', ['A_1'])
        key_store.add_alias('A_3', ['A_1', 'john'])

        for a in ('A_1', 'A_2', 'A_3'):
            match = key_store.get_public_key(a)
            self.assertTrue(generated_pub == match.export_public_key('openssh'))


#################
### Plugins

def validator_test(case, validator, good=[], bad=[]):
    for value in good:
        success, message = validator(value)
        case.assertTrue(success, msg='Validator "{}" should pass for value "{}" instead received: {}'.format(
            validator.__name__,
            value,
            message,
        ))
    for value in bad:
        success, message = validator(value)
        case.assertFalse(success, msg='Validator "{}" should fail for value "{}" instead received: {}'.format(
            validator.__name__,
            value,
            message,
        ))

class SimpleFileLoaderTests(unittest.TestCase):
    test_data = {'foo': 'bar', 'dataz': 42}

    def fake_client(self):
        return object()

    def test_validators(self):
        good_json = json.dumps(self.test_data)

        plugin = commands.SimpleFileLoaders(self.fake_client())
        validators = plugin.filetype_validators

        self.assertIn('json', validators)
        self.assertIn('yaml', validators)
        self.assertIn('yml', validators)
        self.assertEqual(validators['yaml'], validators['yml'])


    def test_json_validator(self):
        plugin = commands.SimpleFileLoaders(self.fake_client())
        json_validator = getattr(plugin, plugin.filetype_validators['json'])

        validator_test(
            self,
            json_validator,
            bad=[None, 'not json'.encode('utf-8'), b''],
            good=[json.dumps(self.test_data).encode('utf-8')],
        )

    def test_yaml_validator(self):
        plugin = commands.SimpleFileLoaders(self.fake_client())
        yaml_validator = getattr(plugin, plugin.filetype_validators['yaml'])

        validator_test(
            self,
            yaml_validator,
            bad=[None, '{not yaml'.encode('utf-8')],
            good=[yaml.dump(self.test_data).encode('utf-8')],
        )



from . import security

def find(items, f):
    for item in items:
        if f(item):
            return item
    return None

class BaseScheme():
    """
        This allows us to toggle the loading from file because I plan on 
        expanding the list of schemes to include variations on 
        padding and/or key size and im not sure if thatll be part of the scheme's
        options or if it should be a separate scheme (perhaps subclass?)

        This class could be used as an "interface" as anything that implements these
        method signatures and returns metadata with a "scheme" property is a valid encx
        scheme.
    """    
    KEY_TYPE_RSA = 'rsa'
    KEY_TYPE_BYTES = 'bytes'

    # Everything but RSA should use the below. We don't need to differentiate them based
    # on size because they're loaded the same and the security classes or scheme classes
    # can validate them later
    key_type = KEY_TYPE_BYTES
    multi_key = False

    def __init__(self, keys, options=None):
        if not keys:
            raise ValueError('All encx encryption schemes require a key!')
        self.keys = keys
        self.options = options 

        if not self.multi_key:
            if len(keys) > 1:
                raise ValueError('Scheme {} does not support multiple keys'.format(
                    self.name,
                ))
            self.key = keys[0]
         
    def encrypt(self, payload):
        raise NotImplemented()

    def decrypt(self, ciphertext, meta):
        raise NotImplemented()

class AESScheme(BaseScheme):
    name = 'AES'
    default_key_size = 16 # In bytes; so 16 is a 128-bit key

    @classmethod
    def generate_key(cls):
        return security.AES.generate_key()

    def encrypt(self, payload):
        encryptor = security.AES(self.key)
        ciphertext, metadata = encryptor.encrypt(payload)
        metadata['scheme'] = self.name
        return ciphertext, metadata

    def decrypt(self, ciphertext, meta):
        iv = meta.get('iv', None)
        if not iv:
            raise ValueError('The AES Security scheme requires an initialization vector')
        decryptor = security.AES(self.key)
        decrypted_payload = decryptor.decrypt(ciphertext, iv)
        return decrypted_payload

class RSAScheme(BaseScheme):
    """
        I'm expecting multiple variations of this, and leaving open the possibility
        of having a single-key scheme that is backwards compatible with my original
        logic. Throwing the core logic in the original class and extending it with
        the needs of the multi-key scheme.
    """
    name = None
    key_type = BaseScheme.KEY_TYPE_RSA

    note_message = 'The payload is encrypted with AES, and the key used is encrypted using RSA'

    @classmethod
    def generate_key(cls):
        return security.RSA(security.RSA.generate_key())

    def _do_encrypt(self, payload, include_aes_key=False):
        # Start by generating an AES key and encrypting the payload
        aes_encryptor = security.AES()
        aes_ciphertext, aes_metadata = aes_encryptor.encrypt(payload)
        aes_key = aes_encryptor.get_key()

        rsa_meta_entries = []
        # Now we use the rsa keys given to encrypt the key we generated to 
        # package with the payload
        for rsa_encryptor in self.keys:
            rsa_ciphertext, rsa_metadata = rsa_encryptor.encrypt(aes_encryptor.key)
            encrypted_key = security.to_b64_str(rsa_ciphertext)
            rsa_meta_entries.append({
                'rsa_meta': rsa_metadata,
                'public_key': rsa_encryptor.export_public_key('openssh'),
                'encrypted_key': encrypted_key,
            })

        meta = {
            'aes': aes_metadata,
            'rsa_entries': rsa_meta_entries,
            'note': self.note_message,
        }
        if include_aes_key:
            return aes_ciphertext, meta, aes_key
        else:
            return aes_ciphertext, meta 

    def _do_decrypt(self, ciphertext, meta):
        rsa_meta_entries = meta.get('rsa_entries', {})
        aes_meta = meta.get('aes', {})

        key = self.keys[0]
        # In-reverse of the encrypt, first we extract the encrypted key from the metadata
        exported_public_key = key.export_public_key('openssh')
        rsa_entry = find(rsa_meta_entries, lambda x: x['public_key'] == exported_public_key)
        if not rsa_entry:
            raise ValueError('No matching RSA key found for decryption')
        encrypted_key_str = rsa_entry.get('encrypted_key', None)
        if not encrypted_key_str:
            raise ValueError('The RSA security scheme requires an encrypted key packaged with payload')
        encrypted_key = security.from_b64_str(encrypted_key_str)

        # Decrypt the AES key that was used on the payload
        rsa_decryptor = key
        aes_key = rsa_decryptor.decrypt(encrypted_key)

        # Now do the decryption of the actual payload
        iv = aes_meta.get('iv', None)
        if not iv:
            raise ValueError('The AES Security scheme requires an initialization vector')
        aes_decryptor = security.AES(aes_key)
        payload = aes_decryptor.decrypt(ciphertext, iv)
        return payload

class RSAMultiKeyScheme(RSAScheme):
    name = 'RSA-AES-mk'
    multi_key = True

    def encrypt(self, payload, include_aes_key=False):
        return self._do_encrypt(payload, include_aes_key=include_aes_key)

    def decrypt(self, ciphertext, meta):
        if len(self.keys) > 1:
            raise ValueError('Decryption operations expect to receive only one key')
        return self._do_decrypt(ciphertext, meta)

all_schemes = [AESScheme, RSAMultiKeyScheme]
schemes = {scheme.name: scheme for scheme in all_schemes}
DEFAULT_SCHEME = RSAMultiKeyScheme

def get_scheme(meta):
    """
        Determine which scheme to use for decryption based on scheme name.
        The complexity of this function should increase as I intend to 
        make variations of schemes based on padding / key types. Im
        not sure how i'll break up the logic, but putting a layer here for
        easy refactoring.
    """
    scheme_name = meta.get('scheme', None)
    if not scheme_name:
        raise ValueError('"scheme" property was not specified. Invalid encx metadata.')
    scheme = schemes.get(scheme_name, None)
    if not scheme:
        raise ValueError('Scheme "{}" is unsupported by this implementation of encx'.format(scheme_name))
    return scheme



from .schemes import DEFAULT_SCHEME, get_scheme, schemes
from .spec import ENCX
from . import security

import yaml

from getpass import getpass
from uuid import uuid4
from pprint import pprint
import logging
import string
import json
import sys
import os
import io

class BasePlugin():
    file_loaders = {}
    filetype_validators = {}
    commands = {}

    def __init__(self, client):
        self.client = client

    def get_config(self, local=True):
        global_config = self.client.get_config()
        if local:
            return global_config.get('plugins', {}).get(self.name, {})
        else:
            return global_config

    def set_config(self, new_configuration, local=True):
        if local:
            config = self.client.get_config()
            if not config.get('plugins', None):
                config['plugins'] = {}
            config['plugins'][self.name] = new_configuration
            self.client.set_config(config)
        else:
            config = new_configuration
            self.client.set_config(config)
        return config

    def finish(self):
        """ Hook to allow post-operation processing (config-saves, etc) """
        pass

class SimpleFileLoaders(BasePlugin):
    name = 'simple_file_loaders'
    file_loaders = {
        '.*': {
            'loader': 'load', # We want to match everything as a fallback
            'writer': 'write', # We want to match everything as a fallback
        }
    }
    filetype_validators = {
        'json': 'json_validator',
        'yaml': 'yaml_validator',
        'yml': 'yaml_validator',
    }

    validation_success_message = 'Everything looks good!'

    def json_validator(self, data):
        try:
            deserialized = json.loads(data.decode('utf-8'))
            return True, self.validation_success_message
        except Exception as e:
            return False, 'Invalid JSON:' + str(e)

    def yaml_validator(self, data):
        try:
            deserialized = yaml.load(data.decode('utf-8'))
            return True, self.validation_success_message
        except Exception as e:
            return False, 'Invalid YAML:' + str(e)

    def load(self, path):
        if not path or path == '-':
            return sys.stdin.buffer.read()
        with open(os.path.expanduser(path), 'rb') as f:
            content = f.read()
        return content

    def write(self, path, data, overwrite=False):
        if path == '-':
            sys.stdout.buffer.write(data)
            return True
        if not overwrite and os.path.exists(path):
            raise FileExistsError()
        with open(os.path.expanduser(path), 'wb') as f:
            f.write(data)

class PluginManagement(BasePlugin):
    name = 'plugin_management'
    commands = {
        'plugin:install': {
            'parser': 'parse_install',
            'run': 'install',
            'help': 'Install a plugin',
        },
        'plugin:list': {
            'run': 'list_plugins',
            'help': 'Show plugins',
        },
        'plugin:uninstall': {
            'parser': 'parse_uninstall',
            'run': 'uninstall',
            'help': 'Uninstall a plugin',
        },
    }

    def _installed_plugin_list(self):
        config = self.get_config(local=False)
        installed_plugins = config.get('installed_plugins', [])
        return installed_plugins

    def parse_install(self, parser):
        parser.add_argument('path', nargs=1, help='Python path to plugin class (e.g. my_module.MyPlugin)')
        parser.add_argument('-f', '--force', action='store_true', help='Uninstall without attempting to load')

    def install(self, args):
        plugin_to_install = args.path.pop()
        if not args.force:
            success, result = self.client.load_plugin(plugin_to_install)
            if not success:
                logging.error('Failed to load plugin {}:'.format(plugin_to_install))
                logging.error(str(result))
                return False
            print('Installing plugin {} ({})'.format(result.name, plugin_to_install))
        else:
            print('Installing plugin {}'.format(result.name))


        installed_plugins = self._installed_plugin_list()
        installed_plugins.append(plugin_to_install)

        config = self.get_config(local=False)
        config['installed_plugins'] = installed_plugins
        self.set_config(config, local=False)
        print('Done')

    def parse_uninstall(self, parser):
        parser.add_argument('name', nargs=1, help='Name of plugin to remove')

    def uninstall(self, args):
        plugin_to_remove = args.name.pop()
        print('Uninstalling plugin {}'.format(plugin_to_remove))
        for plugin_path, plugin in self.client.plugins.items():
            if plugin.name == plugin_to_remove:
                if plugin_path in self.client.base_plugins:
                    print('\tPlugin is a core plugin and cannot be uninstalled!')
                    return False

                # Do removal
                installed_plugins = self._installed_plugin_list()
                installed_plugins.remove(plugin_path)
                config = self.get_config(local=False)
                config['installed_plugins'] = installed_plugins
                self.set_config(config, local=False)
                return True

        print('Plugin not found!')
        return False

    def list_plugins(self, args):
        for plugin_path, plugin in self.client.plugins.items():
            print('{} :: {}'.format(plugin.name, plugin_path))
            for cmd, cmd_info in plugin.commands.items():
                print('\t{}: {}'.format(cmd, cmd_info.get('help', 'No documentation available')))
            print('\n')

class KeyStoreManagement(BasePlugin):
    name = 'keystore_management'
    commands = {
        'keystore:show': {
            'parser': 'parse_show',
            'run': 'show',
            'help': 'Show entries in the keystore',
        },
        'keystore:add_alias': {
            'parser': 'parse_add_alias',
            'run': 'add_alias',
            'help': 'Give one or more keys (or other aliases) an alias',
        },
        'keystore:add_to_alias': {
            'parser': 'parse_add_to_alias',
            'run': 'add_to_alias',
            'help': 'Append to an existing alias one or more keys (or other aliases)',
        },
        'keystore:add_public_key': {
            'parser': 'parse_add_public_keys',
            'run': 'add_public_keys',
            'help': 'Add public key(s) to the store with a given name',
        },
        'keystore:add_public_keys': {
            'parser': 'parse_add_public_keys',
            'run': 'add_public_keys',
            'help': 'Add public key(s) to the store with a given name',
        },
        'keystore:add_private_key': {
            'parser': 'parse_add_private_key',
            'run': 'add_private_key',
            'help': 'Add a private key\'s path to the store giving it a name (key is not copied)',
        },
        'keystore:remove': {
            'parser': 'parse_remove',
            'run': 'remove',
            'help': 'Remove an alias or key from the keystore',
        },
    }

    def parse_show(self, parser):
        parser.add_argument('alias', nargs="?", help='Show only specified alias')

    def show(self, args):
        if args.alias:
            entries = self.client.keystore.get_entries(args.alias)
            print(yaml.dump(entries))
        else:
            print(yaml.dump(self.client.keystore.data))

    def parse_add_alias(self, parser):
        parser.add_argument('alias', nargs=1, help='Name to give to this key/group')
        parser.add_argument('keys', nargs="+", help='Keys or other aliases this should include')

    def add_alias(self, args):
        self.client.keystore.add_alias(args.alias[0], args.keys)

    def parse_add_to_alias(self, parser):
        parser.add_argument('alias', nargs=1, help='Name to give to this key/group')
        parser.add_argument('keys', nargs="+", help='Keys or other aliases this should include')

    def add_to_alias(self, args):
        self.client.keystore.add_to_alias(args.alias[0], args.keys)

    def parse_add_private_key(self, parser):
        parser.add_argument('name', nargs=1, help='Name to give to this key')
        parser.add_argument('path', nargs=1, help='Path to private key')

    def add_private_key(self, args):
        self.client.keystore.add_private_key(args.name[0], args.path[0])

    def parse_add_public_keys(self, parser):
        parser.add_argument('name', nargs=1, help='Name to give to this key')
        parser.add_argument('source', nargs='?', help='Source of public key', default='-')

    def add_public_keys(self, args):
        alias = args.name[0]
        source = args.source
        if source == '-':
            source = sys.stdin.read()
        keys = self.client.get_public_keys([source], use_store=False)
        if not keys:
            logging.error('No public keys found!')
            return False
        if len(keys) == 1:
            self.client.keystore.add_public_key(alias, keys[0])
        else:
            key_names = []
            for i, key in enumerate(keys):
                key_name = '{}_{}'.format(alias, i)
                key_names.append(key_name)
                self.client.keystore.add_public_key(key_name, key)
            self.client.keystore.add_alias(alias, key_names)

    def parse_remove(self, parser):
        parser.add_argument('name', nargs=1, help='Name of key or alias to remove')

    def remove(self, args):
        self.client.keystore.delete_key(args.name[0])

class Encryption(BasePlugin):
    name = 'encryption'
    commands = {
        'set_default_key': {
            'parser': 'parse_set_default_key',
            'run': 'set_default_key',
            'help': 'Set an RSA private key as your default key path',
        },
        'file_meta': {
            'parser': 'parse_file_meta',
            'run': 'file_meta',
            'help': 'View the metadata of an encx file',
        },
        'encrypt': {
            'parser': 'parse_encrypt',
            'run': 'encrypt',
            'help': 'Encrypt into encx format.',
        },
        'decrypt': {
            'parser': 'parse_decrypt',
            'run': 'decrypt',
            'help': 'Decrypt from encx format.',
        },
        'edit': {
            'parser': 'parse_edit',
            'run': 'edit',
            'help': 'Edit a file in encx format. (uses a plaintext tmp file)',
        },
    }

    def parse_set_default_key(self, parser):
        parser.add_argument('key', nargs="?", help='RSA private key')

    def set_default_key(self, args):
        config = self.get_config(local=False)
        if args.key:
            print('Setting default key to {}'.format(args.key))
            config['default_key_path'] = args.key
        else:
            print('Clearing default key')
            config.pop('default_key_path', None)
        self.set_config(config, local=False)

    def parse_decrypt(self, parser):
        parser.add_argument('source', nargs="?", help='A file source')
        parser.add_argument('-k', '--key', dest='key', help='Key to use to decrypt')
        parser.add_argument('-d', '--decode', dest='decode', action='store_true', help='Decode data')
        parser.add_argument('-g', '--signer', default=False, nargs='?', help='Require a signature (optionally: of specified key)')
        parser.add_argument('-a', '--allow_anonymous', action='store_true', help='Allow signing keys not in the keystore (ignored if signing key is specified)')
        parser.add_argument('-x', '--no_verify', action='store_true', help='Do not do any signature verification')

    def decrypt(self, args):
        # If the signer flag is specified (None) but not set to anything, toggle it on
        if args.signer is None:
            signer = True
        else:
            signer = args.signer

        data, metadata = self.client.decrypt_file(
            args.source,
            args.key,
            require_signature=signer,
            allow_anonymous=args.allow_anonymous,
            ignore_signature=args.no_verify,
        )
        if args.decode:
            print(data.decode('utf-8'))
        else:
            sys.stdout.buffer.write(data)

    def parse_encrypt(self, parser):
        parser.add_argument('source', nargs="?", default='-', help='A file source')
        parser.add_argument('-s', '--scheme', dest='scheme', help='Scheme to use to encrypt', default=DEFAULT_SCHEME.name)
        parser.add_argument('-k', '--key', nargs='*', dest='keys', help='Key for encryption', default=None)
        parser.add_argument('-t', '--target', default='-', help='Target path for output (default is "-" for stdout')
        parser.add_argument('-f', '--force', action='store_true', help='Overwrite any existing file')
        parser.add_argument('-g', '--sign', default=False, nargs='?', help='Sign with specified private key')

    def encrypt(self, args):
        source_data = self.client.load_file(args.source)
        if args.scheme not in schemes:
            print('Scheme not found!')
            return False

        # We set the default of args.sign to False in order to be able to know when they 
        # specified "--sign" without a value as we infer that the encryption key includes
        # the private portion and signing should occur with that
        if args.sign is None:
            signer = args.keys[0]
        else:
            signer = args.sign

        scheme = schemes.get(args.scheme)
        self.client.encrypt_file(
            args.target,
            source_data,
            scheme=scheme,
            keys=args.keys,
            overwrite=args.force,
            signer=signer,
        )

    def parse_file_meta(self, parser):
        parser.add_argument('source', nargs="?", help='A file source')
        parser.add_argument('-f', '--format', action='store_true', help='Format for readability')

    def file_meta(self, args):
        metadata = self.client.get_file_meta(args.source)
        if args.format:
            pprint(metadata)
        else:
            print(json.dumps(metadata))

    def parse_edit(self, parser):
        parser.add_argument('source', nargs="?", help='A file source')
        parser.add_argument('-i', '--initialize',
            action='store_true',
            help='Start a new file, or replace existing.',
        )
        parser.add_argument('-f', '--force',
            action='store_true',
            help='Okay to overwrite existing file (when initializing)',
        )
        parser.add_argument('-s', '--scheme',
            help='Scheme to use to encrypt',
            default=DEFAULT_SCHEME.name
        )
        parser.add_argument('-k', '--key', help='Key for encryption')
        parser.add_argument('-r', '--validator', help='Validator to use')

    def edit(self, args):
        # Load from source
        if args.initialize:
            data, metadata = bytes(), {} 
        else:
            data, metadata = self.client.decrypt_file(args.source, args.key)

        # Do edit funtionality
        validator_name, validator = self.client.get_filetype_validator(
            args.validator,
            path=args.source,
        )
        success, result = self.client.edit_data(
            data,
            validator=validator,
            extension=validator_name
        )

        if not success:
            print('Edit failed, no modifications to {} were made'.format(args.source))
            return success

        # Ok write back out to source
        scheme = schemes.get(args.scheme)
        if args.initialize:
            overwrite = bool(args.force) # When initializing you need that extra check
        else:
            overwrite = True # Always overwrite when editing
        self.client.encrypt_file(
            args.source,
            result,
            scheme=scheme,
            key=args.key,
            overwrite=overwrite,
        )

class Keygen(BasePlugin):
    name = 'keygen'
    commands = {
        'keygen': {
            'parser': 'parse_keygen',
            'run': 'keygen',
            'help': 'Generate new encryption keys/passwords/IDs (Sourced by urandom)'
        },
    }

    def parse_keygen(self, parser):
        subparsers = parser.add_subparsers(dest='key_type')
        subparsers.required = True
        byte_key_gen = subparsers.add_parser('key', help='Random bytes generated for a key')
        byte_key_gen.add_argument(
            '-l', '--length',
            dest='length',
            type=int,
            help='length of key in bytes',
            default=16
        )
        byte_key_gen.add_argument(
            '-r', '--raw',
            dest='raw',
            action='store_true',
            help='Dont encode bytes into base64 string',
            default=False
        )
        byte_key_gen.set_defaults(func=self.generate_byte_key)


        rsa_key_gen = subparsers.add_parser(
            'rsa', help='Generate an RSA key in PEM format'
        )
        rsa_key_gen.add_argument(
            '-s', '--size',
            type=int,
            help='length of key in bytes',
            default=2048
        )
        rsa_key_gen.add_argument(
            '-k', '--key',
            help='Output key to file'
        )
        rsa_key_gen.add_argument(
            '-a', '--askpass',
            action='store_true',
            help='Prompt for passphrase',
            default=False
        )
        rsa_key_gen.add_argument(
            '-w', '--passphrase',
            help='Give private key a passphrase',
            default=None
        )
        rsa_key_gen.add_argument(
            '-p', '--public',
            help='Output public key to file'
        )
        rsa_key_gen.set_defaults(func=self.generate_rsa_key)

        uuid_parser = subparsers.add_parser('uuid', help='Random UUID')
        uuid_parser.set_defaults(func=self.generate_uuid)


        random_str_parser = subparsers.add_parser('string', help='Random string')
        random_str_parser.add_argument(
            '-s',
            '--source',
            help='Characters to select from (defaults to alpha-numeric)',
            default=string.ascii_letters + string.digits
        )
        random_str_parser.add_argument(
            '-l',
            '--length',
            type=int,
            help='Length of string',
            default=20
        )
        random_str_parser.set_defaults(func=self.generate_random_string)

    def keygen(self, args):
        args.func(args)

    def generate_byte_key(self, args):
        key_bytes = security.generate_random_bytes(args.length)
        if args.raw:
            sys.stdout.buffer.write(key_bytes)
        else:
            print(security.to_b64_str(key_bytes))

    def generate_uuid(self, args):
        print(security.generate_uuid())

    def generate_random_string(self, args):
        selections = [security.random_choice(args.source) for i in range(args.length)]
        print(''.join(selections))

    def generate_rsa_key(self, args):
        rsa = security.RSA(security.RSA.generate_key())

        if args.askpass:
            passphrase = getpass('Enter passphrase for new key:')
        else:
            passphrase = args.passphrase
        
        private_key = rsa.get_private_key(passphrase=passphrase)
        
        if args.public:
            rsa_pub = rsa.get_public_key()
            with open(args.public, 'wb') as pub_file:
                pub_file.write(rsa_pub)

        if args.key:
            with open(args.key, 'wb') as priv_file:
                priv_file.write(private_key)
        else:
            print(private_key)



import yaml

from . import security
from .spec import ENCX
from .schemes import get_scheme, DEFAULT_SCHEME
from .keystore import KeyStore
from .commands import BasePlugin

from importlib import import_module
from collections import OrderedDict
from pprint import pprint
import requests
import subprocess
import logging
import argparse
import time
import sys
import io
import os
import re

GITHUB_USER_PROTO = 'github://'


class FileLoaderInvalidPath(ValueError):
    pass

class CustomArgParser(argparse.ArgumentParser):
    def error(self, message):
        sys.stderr.write('error: %s\n' % message)
        self.print_help()
        sys.exit(2)

def import_class(path):
    parts = path.split('.')
    module_path = '.'.join(parts[:-1])
    class_name = parts[-1]
    m = import_module(module_path)
    obj = getattr(m, class_name)
    return obj

class EncxClient():
    """ Implements the core of the CLI. Extendable with plugins """

    default_config_dir = os.path.expanduser('~/encx')
    default_config_file = 'encx.yaml'
    default_keystore_file = 'keystore.yaml'
    base_plugins = [
        'encxlib.commands.SimpleFileLoaders',
        'encxlib.commands.PluginManagement',
        'encxlib.commands.KeyStoreManagement',
        'encxlib.commands.Encryption',
        'encxlib.commands.Keygen',
    ]

    def __init__(self, config_path=None):
        self.config_path = config_path
        self._load_configuration()
        self._load_keystore()
        self._load_plugins()
        self.parser = self._build_cli()

    def _load_keystore(self):
        path = self.get_keystore_file_path()
        try:
            contents = security.read_private_path(path)
        except FileNotFoundError as e:
            logging.info('Skipping keystore load as the file appears not to exist')
            self.keystore = KeyStore()
        else:
            data = yaml.safe_load(contents)
            self.keystore = KeyStore(data)
        return self.keystore

    def _load_configuration(self):
        # self.config_path is also a flag to determine whether our
        # config needs to be saved out. So if its falsey i want to
        # keep it that way until i can prove that the default config
        # exists. This prevents the default config path from being
        # created unless explicitly asked for by the user.
        path = self.get_config_file_path()

        # This flag will indicate whether we've made any changes and
        # thus whether we need to save out the file
        self._config_changed = False
        try:
            contents = security.read_private_path(path)
        except FileNotFoundError as e:
            if self.config_path:
                # They specified this file explicitly, warn the user
                logging.warning('Specified config file doesnt exist.')
            # They didnt specify file and  default doesnt exist
            # assume they haven't created it
            self._config = {}
        else:
            # We loaded the file, set the config_path
            self.config_path = path
            self._config = yaml.safe_load(contents)
        return self._config

    def _save_configuration(self, force=False):
        if self._config_changed:
            path = self.get_config_file_path()
            if not os.path.exists(path):
                create_config = input('Create new config at location "{}" (yes/no)?'.format(path))
                if create_config in 'no':
                    return False
            dumped_config = yaml.dump(self._config)
            security.write_private_path(path, dumped_config)
            self._config_changed = False # Reset the flag in case this was forced

        if self.keystore.has_changed():
            keystore_path = self.get_keystore_file_path()
            dumped_data = yaml.dump(self.keystore.export())
            security.write_private_path(keystore_path, dumped_data)

        return True
            

    def _load_plugins(self):
        self.plugins = OrderedDict()
        self.plugins_by_name = OrderedDict()

        # Start with the base plugins
        plugin_paths = self.base_plugins.copy()
        plugin_paths.extend(self.get_config().get('installed_plugins', []))
        for path in plugin_paths:
            success, result = self.load_plugin(path)
            if not success:
                logging.error('Failed to load plugin {}:'.format(path))
                logging.error(str(result))
                continue
            plugin = result(self)

            # Add to client mappings
            self.plugins[path] = plugin
            self.plugins_by_name[result.name] = plugin

    def _build_cli(self):
        self.parser = CustomArgParser(description='encx :: An encryption tool')
        # The following argument will never be used as the global parser consumes
        # it before the arguments get here. However, i still want it in the help
        # message
        self.parser.add_argument('-c', '--config', help="Path to configuration file")

        #############
        ## Commands not in plugins
        subparsers = self.parser.add_subparsers(dest='cmd', parser_class=CustomArgParser)
        subparsers.required = True

        self.commands = {
            'config': self.cmd_show_config
        }
        subparser = subparsers.add_parser('config', help='View configuration')

        ##############
        ## Pull in all commands from plugins
        for plugin_name, plugin in self.plugins.items():
            for command, cmd_options in plugin.commands.items():
                cmd_help = cmd_options.get('help', None)
                parser_name = cmd_options.get('parser', None)
                if parser_name:
                    cmd_parser = getattr(plugin, parser_name)
                else:
                    cmd_parser = None
                cmd_runner = getattr(plugin, cmd_options['run'])
                subparser = subparsers.add_parser(command, help=cmd_help)
                if cmd_parser:
                    cmd_parser(subparser)
                self.commands[command] = cmd_runner

        return self.parser

    def _finish(self):
        """ This is a hook that i'm using to trigger configuration changes """
        self._save_configuration()

        for plugin in self.iterate_plugin_objects():
            if hasattr(plugin, 'finish'):
                plugin.finish()

    ###################
    ### Basic commands
    def cmd_show_config(self, args):
        print(yaml.dump(self.get_config()))

    ### Entry Point ###
    def run_command(self, source):
        args = self.parser.parse_args(source)
        command = args.cmd
        if command is None:
           self.parser.print_help() 
           sys.exit(0)
            
        runner = self.commands.get(command)
        runner(args)
        self._finish()

    ### Plugin API ###
    def force_config_save(self):
        self._save_configuration

    def load_plugin(self, path):
        if path in self.plugins:
            return False, 'Specified plugin is already installed'
        try:
            Plugin = import_class(path)
        except (ImportError, ValueError) as e:
            return False, e
        if not issubclass(Plugin, BasePlugin):

            return False, 'Specified plugin is not a subclass of the encxlib.commands.BasePlugin'
        if Plugin.name in self.plugins_by_name:
            return False, 'Specified plugin name "{}" is already taken'.format(plugin.name)
        return True, Plugin

    def get_config(self):
        return self._config

    def set_config(self, new_config):
        self._config = new_config
        self._config_changed = True

    def default_rsa_key(self):
        return self.get_config().get('default_key_path', None)

    def get_keystore_file_path(self):
        path = os.path.join(self.get_config_dir(), self.default_keystore_file)
        return path

    def get_config_file_path(self):
        path = os.path.join(self.get_config_dir(), self.default_config_file)
        return path

    def get_config_dir(self):
        if self.config_path:
            return os.path.dirname(self.config_path)
        else:
            return self.default_config_dir

    def get_tmp_dir(self):
        if self.config_path:
            return os.path.dirname(self.config_path)
        else:
            return os.path.abspath('.')

    def iterate_plugin_objects(self):
        plugin_list = list(self.plugins.values())
        for plugin in plugin_list[::-1]: # Reverse
            yield plugin

    def load_file(self, path):
        if path is None:
            path = ''
        for plugin in self.iterate_plugin_objects():
            for pattern, loader in plugin.file_loaders.items():
                if re.match(pattern, path):
                    loader = getattr(plugin, loader['loader'])
                    try:
                        return loader(path)
                    except (FileNotFoundError, FileLoaderInvalidPath) as e:
                        continue
        raise FileNotFoundError('Could not find file: {}'.format(path))

    def write_file(self, path, data, overwrite=False):
        for plugin in self.iterate_plugin_objects():
            for pattern, loader in plugin.file_loaders.items():
                writer = getattr(plugin, loader['writer'])
                if re.match(pattern, path):
                    try:
                        return writer(path, data, overwrite=overwrite)
                    except (FileLoaderInvalidPath) as e:
                        continue
        raise ValueError('Could not find a writer for file: {}'.format(path))

    def get_file_meta(self, path):
        source = self.load_file(path)
        encx_file = ENCX.from_file(io.BytesIO(source))
        return encx_file.metadata

    def decrypt_file(self, path, key=None, ignore_signature=False, require_signature=None, allow_anonymous=False):
        loaded_key = self.get_private_key(key, require=True)
        source = self.load_file(path)
        encx_file = ENCX.from_file(io.BytesIO(source))
        Scheme = get_scheme(encx_file.metadata)
        scheme = Scheme(keys=[loaded_key])
        scheme_metadata = encx_file.metadata.get('scheme_metadata', {})
        payload = scheme.decrypt(encx_file.payload.read(), scheme_metadata)

        # Verify signature if necessary
        signature_data = encx_file.metadata.get('signature', None)
        if ignore_signature:
            logging.info('Skipping signature verification')
        elif signature_data:
            signature = signature_data.get('signature', None)
            signing_key_string = signature_data.get('public_key', None)
            signing_key = security.RSA(signing_key_string)
            if not signing_key.verify(payload, signature):
                raise ValueError('Invalid signature!')

            if type(require_signature) != bool:
                # Check that signature matches one of specified 
                match = False
                for key in self.get_public_keys([require_signature]):
                    if key.export_public_key('openssh') == signing_key_string:
                        match = True
                        break
                if not match:
                    raise ValueError('Signing key "{}", does not match specified "{}"'.format(signing_key_string, require_signature))
            elif not self.keystore.is_trusted_key(signing_key) and not allow_anonymous:
                # It should at least be in our store if it isn't specified
                raise ValueError('Signing key {} is not in trusted key store!'.format(signing_key_string))

        elif require_signature:
            raise ValueError('No signature for validating data found!')
        
        return payload, encx_file.metadata

    def encrypt_file(self, path, data, scheme=DEFAULT_SCHEME, keys=None, overwrite=False, signer=None):
        if scheme.key_type == scheme.KEY_TYPE_RSA:
            keys = self.get_public_keys(keys)

        if not keys:
            raise ValueError('No encrypting keys found!')

        scheme_instance = scheme(keys=keys)
        encrypted_payload, scheme_metadata = scheme_instance.encrypt(data)
        metadata = {
            'scheme': scheme.name,
            'scheme_metadata': scheme_metadata,
        }
        if signer:
            signing_key = self.get_private_key(signer)
            if not signing_key:
                raise ValueError('No signing key specified and no default set!')
            signature = signing_key.sign(data)
            metadata['signature'] = {
                'public_key': signing_key.export_public_key('openssh'),
                'signature': signature,
            }

        encx_file = ENCX(metadata, io.BytesIO(encrypted_payload))
        output_stream = encx_file.to_file(io.BytesIO())
        output_stream.seek(0)
        return self.write_file(path, output_stream.read(), overwrite=overwrite)

    def get_filetype_validator(self, name, path=None):
        if not name and not path:
            return None, None
        elif not name:
            # Exclude the encx extension
            parts = path.split('.')
            if parts[-1] == 'encx':
                parts.pop()
            
            # We didnt have a valid extension
            if len(parts) == 1:
                return None, None
            name = parts[-1].lower()

        validator = None
        for plugin in self.iterate_plugin_objects():
            validator_attr = plugin.filetype_validators.get(name, None)
            if validator_attr:
                validator = getattr(plugin, validator_attr)
                break
        return name, validator

    def edit_data(self, data, extension=None, validator=None, timeout=None):
        """
            Create a temporary file to edit the content.  If a valiator
            is given then content is not returned until it passes the validator.

            Polling is used to check for changes to the tmp file which isnt
            ideal but its a minimal delay and is an easy cross-platform check
            that doesn't require a big dependency.
        """
        poll_time = 1
        error_view_time = 2
        success = False
        new_data = None
        if extension is None:
            extension = 'encx-tmp'
        tmp_path = os.path.join(
            self.get_tmp_dir(),
            '{}.{}'.format(security.generate_uuid(), extension),
        )
        tmp_path = os.path.expanduser(tmp_path)
        security.write_private_path(tmp_path, data, mode='wb')
        creation_time = os.path.getmtime(tmp_path)

        while not success:
            try:
                # Launch editor
                subprocess.call(["vim", "-n", tmp_path])

                # Poll FS for change
                if timeout:
                    expire = time.time() + timeout
                else:
                    expire = sys.maxsize # Arbitrarily large number that won't be reached
                modified = False
                print('Waiting for changes to tmp file...(Ctrl + C to cancel)')
                while time.time() < expire:
                    modified_time = os.path.getmtime(tmp_path)
                    if modified_time > creation_time:
                        modified = True
                        break
                    time.sleep(poll_time)

                if not modified:
                    logging.error('Edit timed out! Going to clean up')
                    break

                new_data = security.read_private_path(tmp_path, 'rb')
                if validator:
                    passes, message = validator(new_data)
                    if not passes:
                        logging.error(message)
                        time.sleep(error_view_time)
                        continue # Restart editor, try again
                    else:
                        success = True
                else:
                    success = True

            except KeyboardInterrupt as e:
                logging.error('Cancelling file edit')
                break
            except Exception as e:
                logging.error('Failed to edit file!')
                logging.error(str(e))
                break

        security.remove_path(tmp_path)
        return success, new_data;

    def get_private_key(self, source, require=True):
        if not source:
            match = self.default_rsa_key()
            if not match and require:
                raise ValueError('No private key specified and no default set!')
        elif self.keystore.has_key(source):
            match = self.keystore.get_private_key(source, require_match=False)
        else:
            match = security.load_rsa_key(source)
        if not match and require:
            raise ValueError('No private key matches given source: {}'.format(source))
        return match

    def get_public_keys(self, sources, use_store=True):
        """
            Despite the name, this functions result may be an RSA key object that is
            the private portion (source may be path to private key). The public portion
            needed to encrypt will be derived by the RSA wrapper without additional logic.
        """
        if not sources:
            if self.default_rsa_key():
                return [self.default_rsa_key()]
            else:
                return []

        matches = []
        for source in sources:
            if use_store and self.keystore.has_key(source):
                matches.extend(self.keystore.get_public_keys(source, require_match=False))
                continue

            entries = [source]
            if source.startswith('https://'):
                result = requests.get(source).text
                # Special-casing github's multi-line format
                if result.startswith('ssh-rsa'):
                    entries = result.strip().split('\n')
                else:
                    entries = [result]
            elif source.startswith(GITHUB_USER_PROTO):
                username = source[len(GITHUB_USER_PROTO):]
                result = requests.get('https://github.com/{}.keys'.format(username)).text
                entries = result.strip().split('\n')

            for entry in entries:
                entry_match = security.load_rsa_key(entry)
                if not entry_match:
                    raise ValueError('Source "{}" did not match any keys!'.format(entry))
                matches.append(entry_match)
        return matches



#!/usr/bin/env python

from setuptools import setup

setup(
    name='encx',
    version='0.3.1',
    description='Encryption CLI tool and implementation of the encx file format.',
    author='KJ',
    author_email='jdotpy@users.noreply.github.com',
    url='https://github.com/jdotpy/encx',
    packages=[
        'encxlib',
    ],
    install_requires=[
        'cryptography',
        'pyyaml',
        'requests',
    ],
    scripts=[
        'bin/encx',
    ],
)
